"""Constants for lmdirect"""
ENABLED = "Enabled"
DISABLED = "Disabled"

GW_URL = "https://gw.lamarzocco.io/v1/home/machines"
TOKEN_URL = "https://cms.lamarzocco.io/oauth/v2/token"
CUSTOMER_URL = "https://cms.lamarzocco.io/api/customer"

IP_ADDR = "IP_ADDR"
CLIENT_ID = "CLIENT_ID"
CLIENT_SECRET = "CLIENT_SECRET"
USERNAME = "USERNAME"
PASSWORD = "PASSWORD"